/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <pthread.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>

using namespace std;

#include "cx88_xml_config.h"
#include "cx88_file_capture.h"
#include "cx88_url_capture.h"
#include "cx88_mpeg_driver_factory.h"

#define DEFAULT_XML_CONFIG "~/.cx88.xml"
#define DEFAULT_TUNER_CONFIG "~/.tuner_config"

extern char *optarg;

typedef struct
{
   cx88_driver *driver;
   cx88_xml_config *conf;
   double threshold;
   int socket;
   string str;
} cx88_scan_arg;

static int capture(
   cx88_driver *driver,
   string &channel,
   cx88_xml_config *xmlconf,
   unsigned short port,
   bool auto_start,
   bool interactive,
   unsigned int duration_s,
   uint32_t lock_timeout_ms);

static void print_channel(void *arg, int error);
   
static void capture_complete(void *arg, int error);

static int scan(cx88_scan_arg &arg, cx88_capture_params &orig_params);

static void send_string(int socket, string &str);

int main(int argc, char **argv)
{
   int error = 0, optch = -1;
   char *device = NULL;
   cx88_driver *driver = NULL;
   cx88_xml_config *xmlconf = NULL;
   tuner_config tunerconf;
   capvec captures;
   string channel = CX88_XML_CHANNEL_DEFAULT;
   unsigned int duration_s = 0;
   size_t file_cache = CX88_DEFAULT_FILE_CACHE;
   size_t mtu = CX88_DEFAULT_MTU;
   unsigned short remote_port = 0;
   bool auto_start = true;
   bool interactive = true;
   uint32_t lock_timeout_ms = 1000;
   static const char *help = 
      "Options:\n"
      "\t-d <path>: [Required] Specify device from which to capture\n"
      "\t-c <channel | profile:channel>: [Optional] Specify initial channel, e.g. '7' or 'ATSC:7'\n"
      "\t-u <protocol://hostname:port>: [Optional] URL, Send capture over TCP or UDP to provided host\n"
      "\t-f <filename>: [Optional] Send capture to specified file\n"
      "\t-m <number>: [Optional] Set UDP MTU size (in bytes).  Affects all subsequent -u options until next -m.  Default is 1500 bytes.\n"
      "\t-b <number>: [Optional] Set file capture cache size (in bytes).  Affects all subsequent -f options until next -b.  Default is 4MB.\n"
      "\t-x <filename>: [Optional] Specify XML configuration file; default is "DEFAULT_XML_CONFIG"\n"
      "\t-t <filename>: [Optional] Specify libtuner configuration file; default is "DEFAULT_TUNER_CONFIG"\n"
      "\t-n <value>: [Optional] Terminate capture after specified number of minutes\n"
      "\t-p <value>: [Optional] Specify port on which to listen for remote control connections\n"
      "\t-l: <value> [Optional] Maximum channel lock timeout, in ms\n"
      "\t-q: [Optional] Non-interactive capture--do not listen for input on stdin\n"
      "\t-h: [Optional] Do not automatically start capturing\n"
      "\t-? : Display this message\n";
   while (!error && ((optch = getopt(argc, argv, "d:c:u:f:m:b:x:t:n:p:l:qh?")) != -1))
   {
      switch (optch) 
      {
         case 'd':
            if (device != NULL)
            {
               CX88ERR << "Warning: ignoring duplicate device directive" << endl;
            }
            else if ((device = strdup(optarg)) == NULL)
            {
               CX88ERR << "Unable to store device filename." << endl;
               error = ENOMEM;
            }
            break;
         case 'c':
            try
            {
               channel = optarg;
            }
            catch (...)
            {
               CX88ERR << "Unable to store channel name.";
               error = ENOMEM;
            }
            break;
         case 'u':
         {
            cx88_url_capture *cap = new cx88_url_capture(optarg, error, mtu);
            if (cap == NULL)
            {
               CX88ERR << "Unable to store URL capture object." << endl;
               error = ENOMEM;
            }
            try
            {
               captures.push_back(cap);    
            }
            catch (...)
            {
               delete cap;
               CX88ERR << "Unable to store URL capture object." << endl;
               error = ENOMEM;
            }
            break;
         }
         case 'f':
         {
            cx88_file_capture *cap = new cx88_file_capture(optarg, error, file_cache);
            if (cap == NULL)
            {
               CX88ERR << "Unable to store file capture object." << endl;
               error = ENOMEM;
            }
            try
            {
               captures.push_back(cap);    
            }
            catch (...)
            {
               delete cap;
               CX88ERR << "Unable to store file capture object." << endl;
               error = ENOMEM;
            }
            break;
         }
         case 'm':
            mtu = strtoul(optarg, NULL, 10);
            break;
         case 'b':
            file_cache = strtoul(optarg, NULL, 10);
            break;
         case 'x':
            if (xmlconf != NULL)
            {
               CX88ERR << "Warning: ignoring duplicate cx88 configuration file" << endl;
               break;
            }
            xmlconf = new cx88_xml_config(optarg, tunerconf, error);
            if (xmlconf == NULL)
            {
               CX88ERR << "Unable to store XML configuration object." << endl;
               error = ENOMEM;
            }
            if (error)
            {
               delete xmlconf;
               xmlconf = NULL;
            }
            break;
         case 't':
            error = tunerconf.load_file(optarg);
            break;
         case 'n':
            duration_s = (unsigned long) (strtod(optarg, NULL) * 60);
            break;
         case 'p':
            remote_port = (unsigned short) (strtoul(optarg, NULL, 10));
            break;
         case 'l':
            lock_timeout_ms = (uint32_t) (strtoul(optarg, NULL, 10));
            break;
         case 'q':
            interactive = false;
            break;
         case 'h':
            auto_start = false;
            break;
         case '?':
            cout << help;
            break;
         default:
            break;
      }
   }
   if (!error)
   {
      if (device == NULL)
      {
         CX88ERR << "Error: You must specify a device (-d)." << endl;
         error = EINVAL;
      }
      if (captures.size() == 0)
      {
         CX88ERR << "Error: You must specify at least one capture (-u, -f)." << endl;
         error = EINVAL;
      }
   }
   if (!error && (xmlconf == NULL))
   {
      xmlconf = new cx88_xml_config(DEFAULT_XML_CONFIG, tunerconf, error);
      if (xmlconf == NULL)
      {  
         error = ENOMEM;  
      }
      if (error)
      {
         CX88ERR << "Unable to load cx88 XML configuration file" << endl;
      }
   }
   if (!error)
   {
      if (strstr(device, "cx88mpeg") != NULL)
      {
         driver = cx88_mpeg_driver_factory::create(
            *xmlconf,
            tunerconf,
            device,
            captures,
            error);
         if (!error && (driver == NULL))
         {
            CX88ERR << "Unable to create cx88 MPEG driver" << endl;
            error = ENOMEM;  
         }
      }
      else
      {
         CX88ERR << "Invalid device filename" << endl;
         error = EINVAL;
      }
   }
   if (!error)
   {
      error = capture(driver, channel, xmlconf, remote_port, auto_start, interactive, duration_s, lock_timeout_ms);
   }
   for (capvec::size_type i = 0; i < captures.size(); ++i)
   {
      delete captures[i];  
   }
   delete driver;
   free(device);
   delete xmlconf;
   return error;
}

static void capture_complete(void *arg, int error)
{
   int fd = *(reinterpret_cast <int*> (arg));
   static char c = 'q';
   write(fd, &c, sizeof(c));
}

static void print_channel(void *arg, int error)
{
   cx88_scan_arg * scan_arg = reinterpret_cast<cx88_scan_arg *>(arg);
   double strength = (error ? 0.0 : scan_arg->driver->get_signal_strength());
   if (strength >= scan_arg->threshold)
   {
      ostringstream stream;
      stream << scan_arg->conf->get_channel() << " (" << strength << "%) ";
      cout << stream.str().c_str();
      if (scan_arg->socket != -1)
      {
         stream << endl;
         scan_arg->str += stream.str();
      }
   }
   else
   {
      cout << ". ";  
   }
   if (scan_arg->conf->last_channel())
   {
      send_string(scan_arg->socket, scan_arg->str);
      scan_arg->driver->stop();
   }
   else
   {
      string channel = CX88_XML_CHANNEL_UP;
      if (scan_arg->driver->set_channel(channel) != 0)
      {
         scan_arg->driver->stop();
      }
   }
}

static void send_string(int socket, string &str)
{
   if (socket != -1)
   {
      const char *cstr = str.c_str();
      ssize_t len = (ssize_t)str.size() + 1;
      ssize_t sent = 0;
      while (sent < len)
      {
         ssize_t count = send(socket, cstr, len - sent, 0);
         if (count < 0)
         {
            return;
         }
         cstr += count;
         sent += count;
      }
   }
}

static int capture(
   cx88_driver *driver,
   string &channel,
   cx88_xml_config *xmlconf,
   unsigned short remote_port,
   bool auto_start,
   bool interactive,
   unsigned int duration_s,
   uint32_t lock_timeout_ms)
{
   int error = 0;
   string str = channel;
   cx88_capture_params params;
   params.completion = NULL;
   params.completion_arg = NULL;
   params.channel_event = NULL;
   params.channel_arg = NULL;
   params.duration_s = duration_s;
   params.lock_timeout_ms = lock_timeout_ms;
   fd_set fds;
   int server_sock = -1, client_sock = -1;
   bool halted = !auto_start;
   bool remote_input = false;
   bool started = false;
   int maxfd = 0;
   int controlfds[2];
   controlfds[0] = -1;
   controlfds[1] = -1;
   if (interactive)
   {
      maxfd = fileno(stdin) + 1;  
   }
   if (duration_s != 0)
   {
      if (pipe(controlfds))
      {
         CX88ERR << "Unable to create control file descriptor for scanning: " 
            << strerror(errno) << endl;  
         error = errno;
         goto cleanup;
      }
      if (controlfds[0] >= maxfd)
      {
         maxfd = controlfds[0] + 1;
      }
      params.completion = capture_complete;
      params.completion_arg = &(controlfds[1]);
   }
   if (remote_port != 0)
   {
      struct sockaddr_in server_addr;
      memset(&server_addr, 0, sizeof(server_addr));
      server_addr.sin_family = AF_INET;
      server_addr.sin_port = htons(remote_port);
      server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
      if ((server_sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
      {
         CX88ERR << "Unable to create remote control socket: " << strerror(errno) << endl;
         error = errno;
         goto cleanup;
      }
      if (bind(server_sock, (struct sockaddr*) (&server_addr), sizeof(server_addr)) < 0) 
      {
         close(server_sock);
         CX88ERR << "Unable to bind remote control socket: " << strerror(errno) << endl;
         error = errno;
         goto cleanup;
      }
      if (listen(server_sock, 1) < 0)
      {
         close(server_sock);
         CX88ERR << "Unable to establish socket listener: " << strerror(errno) << endl;
         error = errno;
         goto cleanup;
      }
      if (server_sock >= maxfd)
      {
         maxfd = server_sock + 1;
      }
   }
   if (maxfd == 0)
   {
      CX88ERR << "You must either specify a time limit (-n) or remote control port (-p), or omit the -q option" << endl;
      error = EINVAL;
      goto cleanup;
   }
   if (!halted)
   {
      if ((error = driver->start(channel, params)))
      {
         CX88ERR << "Unable to start driver" << endl;
         goto cleanup;
      }
      started = true;
   }
   int cmd;
   do
   {
      cmd = 0;
      FD_ZERO(&fds);
      if (interactive)
      {
         FD_SET(fileno(stdin), &fds);
      }
      if (server_sock != -1)
      {
         FD_SET(server_sock, &fds);
         if (client_sock != -1)
         {
            FD_SET(client_sock, &fds);
         }
      }
      if (duration_s != 0)
      {
         FD_SET(controlfds[0], &fds);
      }
      remote_input = false;
      if (select(maxfd, &fds, NULL, NULL, NULL) > 0)
      {
         if ((duration_s != 0) && FD_ISSET(controlfds[0], &fds))
         {
            char c = '\0';
            read(controlfds[0], &c, sizeof(c));
            cmd = c;
         }
         else if (FD_ISSET(fileno(stdin), &fds))
         {
            cmd = getchar();
         }
         else if ((server_sock != -1) && (FD_ISSET(server_sock, &fds)))
         {
            struct sockaddr_in client_addr;
            socklen_t client_addr_len;
            if (client_sock != -1)
            {
               close(client_sock);
            }
            client_sock = accept(server_sock, (struct sockaddr *)(&client_addr), &client_addr_len);
            if (client_sock >= maxfd)
            {
               maxfd = client_sock + 1;
            }
            continue;
         }
         else if ((client_sock != -1) && (FD_ISSET(client_sock, &fds)))
         {
            char data = '\0';
            ssize_t count = recv(client_sock, &data, sizeof(data), 0);
            if (count <= 0)
            {
               continue;  
            }
            cmd = data;
            remote_input = true;
         }
      }
      if (remote_input)
      {
         switch (cmd)
         {
            case 'c':
            case 's':
            {
               char c;
               ssize_t count = 0;
               str = "";
               do
               {
                  if ((count = recv(client_sock, &c, sizeof(c), 0)) == 1)
                  {
                     if (c != '\0')
                     {
                        str += c;
                     }
                  }
               } while ((count == 0) || (c != '\0'));
               break;
            }
         }
      }
      else
      {
         switch (cmd)
         {
            case 'c':  
               cout << "Enter channel:" << endl;
               getline(cin, str);
               getline(cin, str);
               break;
            case 's':
               cout << "Enter signal strength threshold:" << endl;
               getline(cin, str);
               getline(cin, str);
               break;
         }
      }
      switch (cmd)
      {
         case 'c':
            if (driver->set_channel(str) != 0)
            {
               CX88ERR << "Unable to change channel" << endl;
            }
            break;
         case 'i':
            if (driver->get_info(str) != 0)
            {
               CX88ERR << "Unable to get driver info" << endl;  
            }
            else
            {
               cout << str.c_str() << endl;
            }
            if (remote_input)
            {
               send_string(client_sock, str);
            }
            break;
         case 's':
         {
            if (halted)
            {
               CX88ERR << "Unable to scan: driver is halted" << endl;
               break;  
            }
            stringstream stream(str);
            cx88_scan_arg arg;
            arg.driver = driver;
            arg.conf = xmlconf;
            arg.socket = (remote_input ? client_sock : -1);
            stream >> arg.threshold;
            scan(arg, params);
            break;  
         }
         case 'h':
            if (!halted)
            {
               driver->stop();  
               halted = true;
            }
            else
            {
               if (!started)
               {
                  str = channel;
               }
               else
               {
                  xmlconf->get_channel_name(str);
               }
               if ((error = driver->start(str, params)))
               {
                  CX88ERR << "Unable to restart driver" << endl;
               }
               halted = false;
               started = true;
            }
            break;
         default:
            break;
      }
   } while (!error && (cmd != 'q'));
   driver->stop();

cleanup:
   if (client_sock != -1)
   {
      close(client_sock);
   }
   if (server_sock != -1)
   {
      close(server_sock);
   }
   if (controlfds[0] != -1)
   {
      close(controlfds[0]);
   }
   if (controlfds[1] != -1)
   {
      close(controlfds[1]);
   }
   return error;
}

int scan(cx88_scan_arg &arg, cx88_capture_params &orig_params)
{
   int error = 0;
   string orig_channel;
   arg.str = "";
   arg.conf->get_channel_name(orig_channel);
   arg.driver->stop();
   int controlfds[2];
   if (pipe(controlfds))
   {
      CX88ERR << "Unable to create control file descriptor for scanning: " 
         << strerror(errno) << endl;  
      return errno;
   }
   cx88_capture_params scan_params;
   scan_params.completion = capture_complete;
   scan_params.completion_arg = &(controlfds[1]);
   scan_params.channel_event = print_channel;
   scan_params.channel_arg = &arg;
   scan_params.duration_s = 0;
   scan_params.lock_timeout_ms = orig_params.lock_timeout_ms;
   fd_set fds;
   FD_ZERO(&fds);
   FD_SET(controlfds[0], &fds);
   cout << "Scanning profile " << arg.conf->get_profile_name() << ' ';
   string start_channel = CX88_XML_CHANNEL_FIRST;
   ofstream temperr("/dev/null");
   streambuf *origerr = cerr.rdbuf(temperr.rdbuf());
   if ((error = arg.driver->start(start_channel, scan_params)))
   {
      CX88ERR << "Unable to start driver for scan" << endl;
      close(controlfds[0]);
      close(controlfds[1]);
      return error;
   }
   char cmd = '\0';
   do
   {
      read(controlfds[0], &cmd, sizeof(cmd));
   } while (cmd != 'q');
   cout << "Finished scan." << endl;
   cerr.rdbuf(origerr);
   close(controlfds[0]);
   close(controlfds[1]);
   if ((error = arg.driver->start(orig_channel, orig_params)))
   {
      CX88ERR << "Unable to restart driver" << endl;
      return error;
   }
   return error;
}
