/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <sys/errno.h>
#include "cx88_driver.h"

cx88_driver::cx88_driver(
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures,
   int &error)
   : m_cx88conf(cx88conf),
     m_tunerconf(tunerconf),
     m_devnode(devnode),
     m_stop(true),
     m_channel_changed(false),
     m_mutex_created(false),
     m_signal_created(false),
     m_captures(captures)
{
   if ((error = pthread_mutex_init(&m_lock, NULL)))
   {
      CX88ERR << "Unable to initialize driver lock" << endl;
      return;  
   }
   m_mutex_created = true;
   if ((error = pthread_cond_init(&m_signal, NULL)))
   {
      CX88ERR << "Unable to initialize driver event signal" << endl;
      return;  
   }
   m_signal_created = true;
}

cx88_driver::~cx88_driver(void)
{
   if (m_signal_created)
   {
      pthread_cond_destroy(&m_signal);
      m_signal_created = false;  
   }
   if (m_mutex_created)
   {
      pthread_mutex_destroy(&m_lock);
      m_mutex_created = false;
   }
}

int cx88_driver::set_channel(const string &channel_string)
{
   cx88_driver_lock lock(this);
   if (!m_stop)
   {
      m_channel_changed = true;
      m_channel = channel_string;
      int error = change_channel(channel_string);
      pthread_cond_signal(&m_signal);
      return error;
   }
   else
   {
      CX88ERR << "Unable to set channel: driver has not been started" << endl;
      return ENXIO;
   }
}

int cx88_driver::start(const string &channel_string, cx88_capture_params &params)
{
   cx88_driver_lock lock(this);
   int error = 0;
   if (m_stop)
   {
      m_params = params;
      m_channel = channel_string;
      m_stop = false;
      if ((error = pthread_create(&m_thread, NULL, thread_routine, this)))
      {
         CX88ERR << "Unable to create driver thread: " << strerror(error) << endl;
         m_stop = true;
         return error;
      }
   }
   return error;
}

int cx88_driver::stop(void)
{
   void *retval = NULL;
   {
      cx88_driver_lock lock(this);
      if (!m_stop)
      {
         m_stop = true;
         pthread_cond_signal(&m_signal);
      }
   }
   pthread_join(m_thread, &retval);
   return (reinterpret_cast <size_t> (retval));
}

void *cx88_driver::thread_routine(void *arg)
{
   cx88_driver *driver = reinterpret_cast<cx88_driver*>(arg);
   int error = driver->capture(driver->m_channel, driver->m_params);
   return ((void*)error);
}

int cx88_driver::start_captures(void)
{
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->start()))
      {
         CX88ERR << "Error starting captures: " << strerror(error) << endl;
         break;
      }
   }
   return error;
}

int cx88_driver::submit_buffer(void *buffer, size_t num_bytes, size_t packet_size)
{
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->submit_buffer(buffer, num_bytes, packet_size)))
      {
         CX88ERR << "Error submitting buffer to captures: " << strerror(error) << endl;
         break;
      }
   }
   return error;
}

void cx88_driver::stop_captures(void)
{
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      m_captures[i]->stop();
   }
}

void cx88_driver::wait_for_event(void)
{
   cx88_driver_lock lock(this);
   if (!m_stop && !m_channel_changed)
   {
      pthread_cond_wait(&m_signal, &m_lock);
   }
}
