/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef __CX88_DRIVER_H__
#define __CX88_DRIVER_H__

#include <pthread.h>
#include "cx88_xml_config.h"
#include "tuner_config.h"

#include <string>
using namespace std;

#include "cx88_capture.h"

typedef void (*cx88_event)(void*, int);
         
typedef struct
{
   cx88_event completion;
   void *completion_arg;
   unsigned int duration_s;
   cx88_event channel_event;
   void *channel_arg;
   uint32_t lock_timeout_ms;
} cx88_capture_params;

class cx88_driver
{
   public:

      cx88_driver(
         cx88_xml_config &cx88conf,
         tuner_config &tunerconf,
         const char *devnode,
         capvec &captures,
         int &error);

      virtual ~cx88_driver(void);

      int start(const string &channel_string, cx88_capture_params &params);
      
      int stop(void);
      
      int set_channel(const string &channel_string);
      
      virtual int get_info(string &info) = 0;
      
      virtual double get_signal_strength(void) = 0;
      
   protected:
      
      cx88_xml_config &m_cx88conf;
      tuner_config &m_tunerconf;
      const char *m_devnode;
      volatile bool m_stop;
      volatile bool m_channel_changed;
      string m_channel;

      int start_captures(void);      
      int submit_buffer(void *buffer, size_t num_bytes, size_t packet_size = 0);
      void stop_captures(void);
      void wait_for_event(void);
      
      virtual int capture(const string &channel_string, cx88_capture_params &params) = 0;
      
      virtual int change_channel(const string &channel_string)
      {
         return 0;
      }
      
   private:
      
      friend class cx88_driver_lock;  
    
      pthread_t m_thread;
      pthread_mutex_t m_lock;
      pthread_cond_t m_signal;
      bool m_mutex_created;
      bool m_signal_created;
      capvec &m_captures;
      cx88_capture_params m_params;
      
      static void *thread_routine(void *arg);
};

class cx88_driver_lock
{
   public:
      
      cx88_driver_lock(cx88_driver *driver)
         : m_driver(driver)
      {
         pthread_mutex_lock(&(m_driver->m_lock));
      }
   
      virtual ~cx88_driver_lock(void)
      {
         pthread_mutex_unlock(&(m_driver->m_lock));
      }
      
   private:
      
      cx88_driver *m_driver;
};

#endif
