/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/errno.h>

#include <sstream>
using namespace std;

#include "cx88_xml_config.h"

cx88_xml_config::cx88_xml_config(const char *filename, tuner_config &tunerconf, int &error)
   : m_file(NULL),
     m_profile(NULL),
     m_range(m_ranges.end()),
     m_channel(0),
     m_alias(m_aliases.end()),
     m_default_channel(CX88_XML_CHANNEL_FIRST),
     m_tunerconf(tunerconf)
{
   m_file = ezxml_parse_file(filename);
   if (m_file == NULL)
   {
      CX88ERR << "Error parsing XML file" << endl;
      error = EINVAL;
      return;
   }
   ezxml_t config = ezxml_child(m_file, CX88_XML_CONFIG_NAME);
   if (config == NULL)
   {
      return;
   }
   if ((error = parse_libtuner(config, m_tunerconf)))
   {
      return;
   }
   ezxml_t alias = ezxml_child(config, CX88_XML_ALIAS_NAME);
   try
   {
      while (alias != NULL)
      {  
         string name(ezxml_attr(alias, CX88_XML_NAME_NAME));
         string channel(ezxml_attr(alias, CX88_XML_CHANNEL_NAME));
         m_aliases.insert(pair<string, string> (name, channel));
         alias = alias->next;     
      }
      ezxml_t default_channel = ezxml_child(config, CX88_XML_CHANNEL_DEFAULT);
      if (default_channel != NULL)
      {
         m_default_channel = default_channel->txt;  
      }
      error = select_channel(m_default_channel);
   }
   catch(...)
   {
      error = ENOMEM;
   }
}

cx88_xml_config::~cx88_xml_config(void)
{
   if (m_file != NULL)
   {
      ezxml_free(m_file);
      m_file = NULL;  
   }
}

int cx88_xml_config::parse_libtuner(ezxml_t parent, tuner_config &conf)
{
   int error = 0;
   ezxml_t libtuner = ezxml_child(parent, CX88_XML_LIBTUNER_NAME);
   while (libtuner != NULL)
   {
      if ((error = conf.load_string(libtuner->txt)))
      {
         return error;
      }
      libtuner = libtuner->next;
   }
   return 0;
}

int cx88_xml_config::select_profile(const char *name)
{
   ezxml_t curprofile = ezxml_child(m_file, CX88_XML_PROFILE_NAME);
   if ((m_profile != NULL) && ((name == NULL) || (strcmp(name, ezxml_attr(m_profile, CX88_XML_NAME_NAME)) == 0)))
   {
      return 0;
   }
   while (curprofile != NULL)
   {
      const char *profile_name = ezxml_attr(curprofile, CX88_XML_NAME_NAME);
      if ((name == NULL) || (strcmp(name, profile_name) == 0))
      {
         m_profile = curprofile;
         try
         {
            m_ranges.clear();
            ezxml_t cur_range = ezxml_child(m_profile, CX88_XML_RANGE_NAME);
            while (cur_range != NULL)
            {
               unsigned long start_chan = 0, end_chan = 0;
               ezxml_t start = ezxml_child(cur_range, CX88_XML_START_NAME);
               ezxml_t end = ezxml_child(cur_range, CX88_XML_END_NAME);
               if (start != NULL)
               {
                  start_chan = strtoul(start->txt, NULL, 10);
               }
               if (end != NULL)
               {
                  end_chan = strtoul(end->txt, NULL, 10);
               }
               cx88_range_t range;
               range.start = start_chan;
               range.end = end_chan;
               range.elem = cur_range;
               int error = parse_libtuner(range.elem, range.tunerconf);
               if (error != 0)
               {
                  return error;
               }
               m_ranges.push_back(range);
               cur_range = cur_range->next;
            }
         }
         catch (...)
         {
            return ENOMEM;
         }
         break;
      }
      curprofile = curprofile->next;  
   }
   m_range = m_ranges.begin();
   if (m_range != m_ranges.end())
   {
      m_channel = m_range->start;
   }
   return 0;
}

const char *cx88_xml_config::get_profile_name(void)
{
   if (m_profile == NULL)
   {
      return NULL;
   }
   return ezxml_attr(m_profile, CX88_XML_NAME_NAME);   
}

int cx88_xml_config::select_channel(unsigned long channel)
{
   if (m_profile == NULL)
   {
      return EINVAL;
   }
   if ((channel >= m_range->start) && (channel <= m_range->end))
   {
      m_channel = channel;
      return 0;  
   }
   for (rangelist::iterator iter = m_ranges.begin(); iter != m_ranges.end(); ++iter)
   {
      if ((iter->start <= channel) && (iter->end >= channel))
      {
         m_tunerconf.remove_config(m_range->tunerconf);
         m_range = iter;
         m_channel = channel;
         return m_tunerconf.add_config(m_range->tunerconf);
      }
   }
   return ENOENT;
}

int cx88_xml_config::select_channel(const string& name)
{
   int error = 0;
   string channel = name;
   if (name == CX88_XML_CHANNEL_UP)
   {
      return channel_up();
   }
   else if (name == CX88_XML_CHANNEL_DOWN)
   {
      return channel_down();
   }
   else if (name == CX88_XML_CHANNEL_DEFAULT)
   {
      channel = m_default_channel;  
   }
   m_alias = m_aliases.find(channel);
   if (m_alias != m_aliases.end())
   {
      channel = m_alias->second;
   }
   string::size_type chan_begin = channel.find_first_of(':', 0);
   string chanstr = channel;
   string profile = channel;
   const char *profile_name = NULL;
   if (chan_begin != string::npos)
   {
      chanstr = channel.substr(chan_begin + 1, channel.length() - chan_begin - 1);
      profile = channel.substr(0, chan_begin);
      profile_name = profile.c_str();
   }
   error = select_profile(profile_name);
   if (!error)
   {
      if (chanstr == CX88_XML_CHANNEL_FIRST)
      {
         return select_first_channel();
      }
      else if (chanstr == CX88_XML_CHANNEL_LAST)
      {
         return select_last_channel();
      }
      error = select_channel(strtoul(chanstr.c_str(), NULL, 10));
   }
   return error;
}

const char *cx88_xml_config::get_property(const char *property)
{
   ezxml_t child = ezxml_child(m_range->elem, property);
   ezxml_t cur_level = m_range->elem;
   while (child == NULL)
   {
      cur_level = cur_level->parent;
      if (cur_level == NULL)
      {
         break;
      }
      child = ezxml_child(cur_level, property);
   }
   if (child != NULL)
   {
      return child->txt;
   }
   return NULL;
}

int cx88_xml_config::parse_modulation(const char *mod_string, dvb_modulation_t &modulation)
{
   int error = 0;  
   if (strcasecmp(mod_string, "VSB8") == 0)
   {
      modulation = DVB_MOD_VSB_8; 
   }
   else if (strcasecmp(mod_string, "VSB16") == 0)
   {
      modulation = DVB_MOD_VSB_16;
   }
   else if (strcasecmp(mod_string, "QAM16") == 0)
   {
      modulation = DVB_MOD_QAM_16;
   }
   else if (strcasecmp(mod_string, "QAM32") == 0)
   {
      modulation = DVB_MOD_QAM_32;
   }
   else if (strcasecmp(mod_string, "QAM64") == 0)
   {
      modulation = DVB_MOD_QAM_64;
   }
   else if (strcasecmp(mod_string, "QAM128") == 0)
   {
      modulation = DVB_MOD_QAM_128;
   }
   else if (strcasecmp(mod_string, "QAM256") == 0)
   {
      modulation = DVB_MOD_QAM_256;
   }
   else if (strcasecmp(mod_string, "QAM") == 0)
   {
      modulation = DVB_MOD_QAM_AUTO;
   }
   else if (strcasecmp(mod_string, "QPSK") == 0)
   {
      modulation = DVB_MOD_QPSK;
   }
   else if (strcasecmp(mod_string, "OFDM") == 0)
   {
      modulation = DVB_MOD_OFDM;
   }
   else
   {
      CX88ERR << "Unrecognized DVB modulation type: " << modulation << endl;
      error = EINVAL;
   }
   return error;  
}

int cx88_xml_config::parse_fec(const char *fec_string, dvb_fec_t &fec)
{
   int error = 0;
   if (strcasecmp(fec_string, "none") == 0)
   {
      fec = DVB_FEC_NONE; 
   }
   else if (strcmp(fec_string, "1/2") == 0)
   {
      fec = DVB_FEC_1_2;
   }
   else if (strcmp(fec_string, "2/3") == 0)
   {
      fec = DVB_FEC_2_3;
   }
   else if (strcmp(fec_string, "3/4") == 0)
   {
      fec = DVB_FEC_3_4;
   }
   else if (strcmp(fec_string, "4/5") == 0)
   {
      fec = DVB_FEC_4_5;
   }
   else if (strcmp(fec_string, "5/6") == 0)
   {
      fec = DVB_FEC_5_6;
   }
   else if (strcmp(fec_string, "6/7") == 0)
   {
      fec = DVB_FEC_6_7;
   }
   else if (strcmp(fec_string, "7/8") == 0)
   {
      fec = DVB_FEC_7_8;
   }
   else if (strcmp(fec_string, "8/9") == 0)
   {
      fec = DVB_FEC_8_9;
   }
   else if (strcasecmp(fec_string, "auto") != 0)
   {
      CX88ERR << "Unrecognized code rate type: " << fec_string << endl;
      error = EINVAL;
   }
   return error;
}

int cx88_xml_config::parse_qam(dvb_qam_settings &qam)
{
   int error = 0;
   qam.symbol_rate = DVB_SYMBOL_RATE_AUTO;
   qam.code_rate = DVB_FEC_AUTO;
   const char *symbol_rate = get_property(CX88_XML_SYMBOL_RATE);
   if (symbol_rate != NULL)
   {
      qam.symbol_rate = (uint32_t)(strtoul(symbol_rate, NULL, 10));
   }
   const char *fec = get_property(CX88_XML_CODE_RATE);
   if (fec != NULL)
   {
      error = parse_fec(fec, qam.code_rate);
   }
   return error;
}

int cx88_xml_config::parse_ofdm(dvb_ofdm_settings &ofdm)
{
   int error = 0;
   ofdm.subcarrier_modulation = DVB_MOD_UNKNOWN;
   ofdm.high_prio_code_rate = DVB_FEC_AUTO;
   ofdm.low_prio_code_rate = DVB_FEC_AUTO;
   ofdm.subcarrier_mode = DVB_OFDM_AUTO_SUBCARRIERS;
   ofdm.guard_interval = DVB_OFDM_GUARD_AUTO;
   ofdm.hierarchy = DVB_OFDM_HIERARCHY_AUTO;
   const char *submodulation = get_property(CX88_XML_SUBMODULATION);
   if (submodulation != NULL)
   {
      error = parse_modulation(submodulation, ofdm.subcarrier_modulation);  
   }
   const char *hp_code_rate = get_property(CX88_XML_CODE_RATE_HP);
   if (!error && (hp_code_rate != NULL))
   {
      error = parse_fec(hp_code_rate, ofdm.high_prio_code_rate);
   }
   const char *lp_code_rate = get_property(CX88_XML_CODE_RATE_LP);
   if (!error && (lp_code_rate != NULL))
   {
      error = parse_fec(lp_code_rate, ofdm.low_prio_code_rate);
   }
   const char *subcarriers = get_property(CX88_XML_SUBCARRIERS);
   if (!error && (subcarriers != NULL))
   {
      if (strcasecmp(subcarriers, "2k") == 0)
      {
         ofdm.subcarrier_mode = DVB_OFDM_2K_SUBCARRIERS;
      }
      else if (strcasecmp(subcarriers, "4k") == 0)
      {
         ofdm.subcarrier_mode = DVB_OFDM_4K_SUBCARRIERS;
      }
      else if (strcasecmp(subcarriers, "8k") == 0)
      {
         ofdm.subcarrier_mode = DVB_OFDM_8K_SUBCARRIERS;
      }
      else if (strcasecmp(subcarriers, "auto") != 0)
      {
         CX88ERR << "Unrecognized subcarrier mode: " << subcarriers << endl;
         return EINVAL;
      }
   }
   const char *guard_interval = get_property(CX88_XML_GUARD_INTERVAL);
   if (!error && (guard_interval != NULL))
   {
      if (strcmp(guard_interval, "1/32") == 0)
      {
         ofdm.guard_interval = DVB_OFDM_GUARD_1_32;
      }
      else if (strcmp(guard_interval, "1/16") == 0)
      {
         ofdm.guard_interval = DVB_OFDM_GUARD_1_16;
      }
      else if (strcmp(guard_interval, "1/8") == 0)
      {
         ofdm.guard_interval = DVB_OFDM_GUARD_1_8;
      }
      else if (strcmp(guard_interval, "1/4") == 0)
      {
         ofdm.guard_interval = DVB_OFDM_GUARD_1_4;  
      }
      else if (strcasecmp(guard_interval, "auto") != 0)
      {
         CX88ERR << "Unrecognized guard interval: " << guard_interval << endl;
         return EINVAL;
      }
   }
   const char *hierarchy = get_property(CX88_XML_HIERARCHY);
   if (!error && (hierarchy != NULL))
   {
      if (strcasecmp(hierarchy, "none") == 0)
      {
         ofdm.hierarchy = DVB_OFDM_HIERARCHY_NONE;
      }
      else
      {
         switch (strtoul(hierarchy, NULL, 10))
         {
            case 1:
               ofdm.hierarchy = DVB_OFDM_HIERARCHY_1;
               break;
            case 2:
               ofdm.hierarchy = DVB_OFDM_HIERARCHY_2;
               break;
            case 4:
               ofdm.hierarchy = DVB_OFDM_HIERARCHY_4;
               break;
            default:  
               CX88ERR << "Unrecognized hierarchy value: " << hierarchy << endl;
               return EINVAL;
         }
      }
   }
   return error;
}

int cx88_xml_config::select_dvb_channel(const string &channel, dvb_channel &dvbchan)
{
   int error = select_channel(channel);
   if (!error)
   {
      const char *bandwidth = get_property(CX88_XML_BANDWIDTH_NAME);
      const char *base_freq = get_property(CX88_XML_FREQ_NAME);
      const char *modulation = get_property(CX88_XML_MODULATION_NAME);
      if ((bandwidth != NULL) && (base_freq != NULL) && (modulation != NULL))
      {
         dvbchan.bandwidth_hz = (uint32_t) (strtoul(bandwidth, NULL, 10));
         dvbchan.frequency_hz = (uint64_t)
            (strtoull(base_freq, NULL, 10) + ((m_channel - m_range->start) * dvbchan.bandwidth_hz));
         error = parse_modulation(modulation, dvbchan.modulation);
         if (!error)
         {
            if ((dvbchan.modulation >= DVB_MOD_QAM_16) && (dvbchan.modulation <= DVB_MOD_QAM_AUTO))
            {
               error = parse_qam(dvbchan.modulation_settings.qam);
            }
            else if (dvbchan.modulation == DVB_MOD_QPSK)
            {
               error = parse_qam(dvbchan.modulation_settings.qpsk);
            }
            else if (dvbchan.modulation == DVB_MOD_OFDM)
            {
               error = parse_ofdm(dvbchan.modulation_settings.ofdm);
            }
         }
      }
      else
      {
         CX88ERR << "Bandwidth, base frequency, or modulation not specified" << endl;
         return EINVAL;
      }
      dvbchan.inversion = DVB_INVERSION_AUTO;
      const char *inversion = get_property(CX88_XML_INVERSION);
      if (!error && (inversion != NULL))
      {
         if (strcasecmp(inversion, "off") == 0)
         {
            dvbchan.inversion = DVB_INVERSION_OFF;    
         }
         else if (strcasecmp(inversion, "on") == 0)
         {
            dvbchan.inversion = DVB_INVERSION_ON;
         }
         else if (strcasecmp(inversion, "auto") != 0)
         {
            CX88ERR << "Invalid spectral inversion setting: " << inversion << endl;
            return EINVAL;  
         }
      }
   }
   return error;
}

int cx88_xml_config::channel_up(void)
{
   if (m_alias != m_aliases.end())
   {
      ++m_alias;
      if (m_alias == m_aliases.end())
      {
         m_alias = m_aliases.begin();    
      }
      return select_channel(m_alias->first);
   }
   if (m_range == m_ranges.end())
   {
      return ENOENT;
   }
   else if (m_channel < m_range->end)
   {
      ++m_channel;
      return 0;
   }
   else
   {
      m_tunerconf.remove_config(m_range->tunerconf);
      ++m_range;
      if (m_range == m_ranges.end())
      {
         m_range = m_ranges.begin();
      }
      m_channel = m_range->start;
      return m_tunerconf.add_config(m_range->tunerconf);
   }
}

int cx88_xml_config::channel_down(void)
{
   rangelist::iterator range = m_range;
   if (m_alias != m_aliases.end())
   {
      if (m_alias == m_aliases.begin())
      {
         m_alias = m_aliases.end();
      }
      --m_alias;
      return select_channel(m_alias->first);
   }
   if (m_range == m_ranges.end())
   {
      return ENOENT;
   }
   else if (m_channel > m_range->start)
   {
      --m_channel;
      return 0;
   }
   else if (m_range == m_ranges.begin())
   {
      m_range = m_ranges.end();
   }
   m_tunerconf.remove_config(range->tunerconf);
   --m_range;
   m_channel = m_range->end;
   return m_tunerconf.add_config(m_range->tunerconf);
}

int cx88_xml_config::select_first_channel(void)
{
   if (m_range == m_ranges.end())
   {
      return ENOENT;
   }
   rangelist::iterator range = m_range;
   m_range = m_ranges.begin();
   m_channel = m_range->start;
   if (range != m_range)
   {
      m_tunerconf.remove_config(range->tunerconf);
      return m_tunerconf.add_config(m_range->tunerconf);  
   }
   return 0;
}
      
int cx88_xml_config::select_last_channel(void)
{
   if (m_range == m_ranges.end())
   {
      return ENOENT;
   }
   rangelist::iterator range = m_range;
   m_range = m_ranges.end();
   --m_range;
   m_channel = m_range->end;
   if (range != m_range)
   {
      m_tunerconf.remove_config(range->tunerconf);
      return m_tunerconf.add_config(m_range->tunerconf);  
   }
   return 0;
}

void cx88_xml_config::get_channel_name(string &name)
{
   if (m_alias != m_aliases.end())
   {
      name = m_alias->first;
   }
   else
   {      
      ostringstream stream;
      stream << get_profile_name() << ':' << get_channel();
      name = stream.str();
   }
}

